/*
 * Copyright 2011 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.devtools.j2objc.gen;

import com.google.common.io.Files;
import com.google.devtools.j2objc.Options;
import com.google.devtools.j2objc.util.ErrorUtil;
import com.google.devtools.j2objc.util.NameTable;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.PackageDeclaration;

import java.io.File;
import java.io.IOException;

/**
 * This class handles common actions shared by the header, implementation, and
 * statement generators.
 *
 * @author Tom Ball
 */
public abstract class SourceFileGenerator {
  private final SourceBuilder builder;
  private final CompilationUnit unit;
  private final File outputDirectory;
  private final String sourceFileName;
  private final String source;

  public SourceFileGenerator(String sourceFileName, String source, CompilationUnit unit, boolean emitLineDirectives) {
    builder = new SourceBuilder(unit, sourceFileName, source, emitLineDirectives);
    this.sourceFileName = sourceFileName;
    this.source = source;
    this.unit = unit;
    this.outputDirectory = Options.getOutputDirectory();
  }

  /**
   * Returns the name this source file should be saved as.  For example,
   * foo/bar/Mumble.java translates to $(OUTPUT_DIR)/foo/bar/Mumble.m for
   * Objective-C implementation files.  If --no-package-directories is
   * specified, though, the output file is $(OUTPUT_DIR)/Mumble.m.
   * <p>
   * Note: class names are still camel-cased to avoid name collisions.
   */
  protected String getOutputFileName(CompilationUnit node) {
    String javaName = NameTable.getMainJavaName(node, sourceFileName);
    PackageDeclaration pkg = node.getPackage();
    if (Options.usePackageDirectories() || pkg == null) {
      return javaName.replace('.', File.separatorChar) + getSuffix();
    } else {
      String pkgName = pkg.getName().getFullyQualifiedName();
      return javaName.substring(pkgName.length() + 1) + getSuffix();
    }
  }

  /**
   * Returns the suffix for files created by this generator.
   */
  protected abstract String getSuffix();

  /**
   * Returns true if a native method has an OCNI block.
   */
  protected boolean hasNativeCode(MethodDeclaration m) {
    return hasNativeCode(m, false);
  }

  /**
   * Returns true if a native method has an OCNI block, warning if JSNI
   * delimiters are found instead.
   */
  protected boolean hasNativeCode(MethodDeclaration m, boolean reportJsniWarnings) {
    assert (m.getModifiers() & Modifier.NATIVE) > 0;
    String nativeCode = extractNativeCode(m.getStartPosition(), m.getLength(), reportJsniWarnings);
    return nativeCode != null;
  }

  /**
   * Returns text from within a source code range, where that text is
   * surrounded by OCNI-like tokens ("/&#42;-[" and "]-&#42;/").
   */
  protected String extractNativeCode(int offset, int length) {
    return extractNativeCode(offset, length, true);
  }

  /**
   * Returns text from within a source code range, where that text is
   * surrounded by OCNI-like tokens ("/&#42;-[" and "]-&#42;/"), warning
   * if JSNI delimiters are found instead.
   *
   * @param offset the offset into the source to begin searching for
   *     a JSNI region
   * @param length the length of the text range to search
   * @param reportWarnings if true, warn that JSNI delimiters were found
   * @return the extracted text between the OCNI delimiters, or null if
   *     a pair of JSNI delimiters aren't in the specified text range
   */
  protected String extractNativeCode(int offset, int length, boolean reportWarnings) {
    String text = source.substring(offset, offset + length);
    int start = text.indexOf("/*-[");  // start after the bracket
    int end = text.lastIndexOf("]-*/");

    if (start == -1 || end <= start) {
      if (reportWarnings && Options.jsniWarnings()) {
        start = text.indexOf("/*-{");
        end = text.lastIndexOf("}-*/");
        if (start != -1 && end > start) {
          String message = String.format("JSNI comment found: %s:%d",
              sourceFileName, builder.getLineNumber(offset));
          ErrorUtil.warning(message);
        }
      }
      return null;
    }
    return text.substring(start + 4, end);
  }

  protected void save(String path) {
    try {
      File outputFile = new File(outputDirectory, path);
      File dir = outputFile.getParentFile();
      if (dir != null && !dir.exists()) {
        if (!dir.mkdirs()) {
          ErrorUtil.warning("cannot create output directory: " + outputDirectory);
        }
      }
      String source = builder.toString();

      // Make sure file ends with a new-line.
      if (!source.endsWith("\n")) {
        source += '\n';
      }

      Files.write(source, outputFile, Options.getCharset());
    } catch (IOException e) {
      ErrorUtil.error(e.getMessage());
    } finally {
      reset();
    }
  }

  protected void print(String s) {
    builder.print(s);
  }

  protected void print(char c) {
    builder.print(c);
  }

  protected void printf(String format, Object... args) {
    builder.printf(format, args);
  }

  protected void println(String s) {
    builder.println(s);
  }

  protected void newline() {
    builder.newline();
  }

  protected void indent() {
    builder.indent();
  }

  protected void unindent() {
    builder.unindent();
  }

  protected void printIndent() {
    builder.printIndent();
  }

  protected char[] pad(int n) {
    return builder.pad(n);
  }

  protected void reset() {
    builder.reset();
  }

  protected void syncLineNumbers(ASTNode node) {
    builder.syncLineNumbers(node);
  }

  protected void printStart(String path) {
    builder.printStart(path);
  }

  protected String reindent(String code) {
    return builder.reindent(code);
  }

  protected CompilationUnit getUnit() {
    return unit;
  }

  protected SourceBuilder getBuilder() {
    return builder;
  }

  protected String getSourceFileName() {
    return sourceFileName;
  }
}
